
local DAMAGE_TAKEN_MEMORY_LENGTH = 4
local EARLY_CAST_EXPIRE_MARGIN = 3
local MINIMUM_TIME_BETWEEN_INTERRUPTS = .3

local lastInterrupt = 0
local checkResults
local inCombat = false
local lastBlockableAttackTaken = 0
local lastUnblockableAttackTaken = 0
local enemyCount = 0
local powerGeneration = {
	["Revenge"] = 15,
	["Shield Slam"] = 20,
	["Berserker Rage"] = 10,
	["Battle Shout"] = 20,
	["Commanding Shout"] = 20,
	["Sword and Board"] = 5,
}


function WhatsNext_GetPowerGeneration(spellName)
	local rtnAmount = powerGeneration[spellName] or 0
	if spellName == "Shield Slam" and UnitBuff("player", "Sword and Board") then
		rtnAmount = rtnAmount + powerGeneration["Sword and Board"] 
	end
	
	return rtnAmount
end


local checks = {
	["Minimum Power"] = {
		test = function(checkTable) 
			return UnitPower("player") >= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's power is at or above " .. tostring(checkTable.amount) .. "."
		end,
		arguments = {
			["amount"] = "integer"
		}
	},
	["Excess Power"] = {
		test = function(checkTable) 
			return UnitPowerMax("player") < UnitPower("player") + WhatsNext_GetPowerGeneration(checkTable.spellName)
		end,
		text = function(checkTable) 
			return tostring(checkTable.spellName) .. " would generate excess power."
		end,
		arguments = {
			["spellName"] = "text"
		}
	},
	["Maximum Power"] = {
		test = function(checkTable) 
			return UnitPower("player") <= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's power is at or below " .. tostring(checkTable.amount) .. "."
		end,
		arguments = {
			["amount"] = "integer"
		}
	},
	["Minimum Missing Power"] = {
		test = function(checkTable) 
			return UnitPowerMax("player") - UnitPower("player") >= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's power is at least " .. tostring(checkTable.amount) .. " below their maximum power."
		end,
		arguments = {["amount"] = "integer"}
	},
	["Maximum Missing Power"] = {
		test = function(checkTable) 
			return UnitPowerMax("player") - UnitPower("player") <= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's power is no more than " .. tostring(checkTable.amount) .. " below their maximum power."
		end,
		arguments = {["amount"] = "integer"}
	},
	["Minimum Resultant Power"] = {
		test = function(checkTable) 
			local _, _, _, cost, _, _, _, _, _ = GetSpellInfo(checkTable.spellName)
			return UnitPower("player") - cost >= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's power would be at least " .. tostring(checkTable.amount) .. " after casting ".. tostring(checkTable.spellName) .."."
		end,
		arguments = {["amount"] = "integer",["spellName"] = "text"}
	},
	["Maximum Resultant Power"] = {
		test = function(checkTable) 
			local _, _, _, cost, _, _, _, _, _ = GetSpellInfo(checkTable.spellName)
			return UnitPowerMax("player") - cost <= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's power would be no more than " .. tostring(checkTable.amount) .. " after casting ".. tostring(checkTable.spellName) .."."
		end,
		arguments = {["amount"] = "integer",["spellName"] = "text"}
	},
	["Minimum Cost"] = {
		test = function(checkTable) 
			local _, _, _, cost, _, _, _, _, _ = GetSpellInfo(checkTable.spellName)
			return cost >= checkTable.amount
		end,
		text = function(checkTable) 
			return tostring(checkTable.spellName) .. " costs at least " .. tostring(checkTable.amount) .. " power."
		end,
		arguments = {["amount"] = "integer",["spellName"] = "text"}
	},
	["Maximum Cost"] = {
		test = function(checkTable) 
			local _, _, _, cost, _, _, _, _, _ = GetSpellInfo(checkTable.spellName)
			return cost <= checkTable.amount
		end,
		text = function(checkTable) 
			return tostring(checkTable.spellName) .. " costs no more than " .. tostring(checkTable.amount) .. " power."
		end,
		arguments = {["amount"] = "integer",["spellName"] = "text"}
	},
	["Target Is Interruptable"] = {
		test = function(checkTable) 
			local spell1, _, _, _, _, _, _, _, notInterruptible1  = UnitCastingInfo("target")
			local spell2, _, _, _, _, _, _, notInterruptible2  = UnitChannelInfo("target")
			return (spell1 and not notInterruptible1)  or (spell2 and not notInterruptible2)
		end,
		text = function(checkTable) 
			return "target is casting an interruptable spell."
		end,
		arguments = {}
	},
	["Tanking"] = {
		test = function(checkTable) 
			local status = UnitThreatSituation("player")
			return status and status > 1
		end,
		text = function(checkTable) 
			return "player is tanking at least 1 enemy NPC."
		end,
		arguments = {}
	},
	["Tanking Target"] = {
		test = function(checkTable) 
			local status = UnitThreatSituation("player")
			return status and status > 1
		end,
		text = function(checkTable) 
			return "player is tanking the target NPC."
		end,
		arguments = {}
	},
	["Player Crowd Controlled"] = {
		test = function(checkTable) 
			return HasFullControl() ~= 1
		end,
		text = function(checkTable) 
			return "player is suffering from some form of crowd control."
		end,
		arguments = {}
	},
	["Target Need Debuff"] = {
		test = function(checkTable) 
			local name, _, _, _, _, _, expirationTime  = UnitDebuff("target", checkTable.debuffName)
			return UnitName("target") and (not name or expirationTime - EARLY_CAST_EXPIRE_MARGIN < GetTime())
		end,
		text = function(checkTable) 
			return "target is missing " .. tostring(checkTable.debuffName) .. " or it will drop off within " .. EARLY_CAST_EXPIRE_MARGIN .. " seconds." 
		end,
		arguments = {["debuffName"] = "text"}
	},
	["In Range"] = {
		test = function(checkTable) 
			return UnitName("target") and IsSpellInRange(checkTable.spellName,"target") == 1
		end,
		text = function(checkTable) 
			return "target is within range of " .. tostring(checkTable.spellName) .. "."
		end,
		arguments = {["spellName"] = "text"}
	},
	["Target In 9.9 Yards"] = {
		test = function(checkTable) 
			return UnitName("target") and CheckInteractDistance("target", 3)
		end,
		text = function(checkTable) 
			return "target is within 9.9 yards."
		end,
		arguments = {}
	},
	["Target In 11.11 Yards"] = {
		test = function(checkTable) 
			return UnitName("target") and CheckInteractDistance("target", 2)
		end,
		text = function(checkTable) 
			return "target is within 11.11 yards."
		end,
		arguments = {}
	},
	["Target In 28 Yards"] = {
		test = function(checkTable) 
			return UnitName("target") and CheckInteractDistance("target", 1)
		end,
		text = function(checkTable) 
			return "target is within 28 yards."
		end,
		arguments = {}
	},
	["In Combat"] = {
		test = function(checkTable) 
			return inCombat
		end,
		text = function(checkTable) 
			return "player is in combat."
		end,
		arguments = {}
	},
	["Taking Damage"] = {
		test = function(checkTable) 
			return GetTime() - math.max(lastBlockableAttackTaken,lastUnblockableAttackTaken) <= DAMAGE_TAKEN_MEMORY_LENGTH
		end,
		text = function(checkTable) 
			return "player has taken damage from an enemy unit in the last " .. DAMAGE_TAKEN_MEMORY_LENGTH .. " seconds."
		end,
		arguments = {}
	},
	["Taking Blockable Damage"] = {
		test = function(checkTable) 
			return GetTime() - lastBlockableAttackTaken <= DAMAGE_TAKEN_MEMORY_LENGTH
		end,
		text = function(checkTable) 
			return "player has taken blockable damage from an enemy unit in the last " .. DAMAGE_TAKEN_MEMORY_LENGTH .. " seconds."
		end,
		arguments = {}
	},
	["Taking Unblockable Damage"] = {
		test = function(checkTable) 
			return GetTime() - lastUnblockableAttackTaken <= DAMAGE_TAKEN_MEMORY_LENGTH
		end,
		text = function(checkTable) 
			return "player has taken unblockable damage from an enemy unit in the last " .. DAMAGE_TAKEN_MEMORY_LENGTH .. " seconds."
		end,
		arguments = {}
	},
	["Maximum Health Percent"] = {
		test = function(checkTable) 
			return UnitHealth("player")/UnitHealthMax("player") <= checkTable.amount
		end,
		text = function(checkTable) 
			return "player's health percent is less then or equal to " .. tostring(checkTable.amount) .. "."
		end,
		arguments = {["amount"] = "percent"}
	},
	["Target Maximum Health Percent"] = {
		test = function(checkTable) 
			return UnitHealth("target")/UnitHealthMax("target") <= checkTable.amount
		end,
		text = function(checkTable) 
			return "target's health percent is less then or equal to " .. tostring(checkTable.amount) .. "."
		end,
		arguments = {["amount"] = "percent"}
	},
	["Maximum Enemy Count"] = {
		test = function(checkTable) 
			return enemyCount <= checkTable.amount
		end,
		text = function(checkTable) 
			return "exists " .. tostring(checkTable.amount) .. " or fewer known enemies."
		end,
		arguments = {["amount"] = "integer"}
	},
	["Minimum Enemy Count"] = {
		test = function(checkTable) 
			return enemyCount >= checkTable.amount
		end,
		text = function(checkTable) 
			return "exists " .. tostring(checkTable.amount) .. " or more known enemies."
		end,
		arguments = {["amount"] = "integer"}
	},
	["Attackable Target"] = {
		test = function(checkTable) 
			return UnitName("target") and UnitCanAttack("player","target") and not UnitIsDeadOrGhost("target") 
		end,
		text = function(checkTable) 
			return "target can be attacked."
		end,
		arguments = {}
	},
	["In Boss Fight"] = {
		test = function(checkTable) 
			return UnitExists("boss1") or UnitExists("boss2") or UnitExists("boss3") or UnitExists("boss4")
		end,
		text = function(checkTable) 
			return "player is in combat with an boss enemy unit."
		end,
		arguments = {}
	},
	["Target Is"] = {
		test = function(checkTable) 
			return UnitName("target") == checkTable.name
		end,
		text = function(checkTable) 
			return "target's name is ".. tostring(checkTable.name) .. "."
		end,
		arguments = {["name"] = "text"}
	},
	["Player Buffed"] = {
		test = function(checkTable) 
			return UnitBuff("player", checkTable.buffName) 
		end,
		text = function(checkTable) 
			return "player has the " .. tostring(checkTable.buffName) .. " buff."
		end,
		arguments = {buffName = "text"}
	},
	["Player Buffed By Player"] = {
		test = function(checkTable) 
			local name, _, _, _, _, _, _, unit = UnitBuff("player", checkTable.buffName) 
			return name and UnitIsUnit(unit,"player")
		end,
		text = function(checkTable) 
			return "player has their own " .. tostring(checkTable.buffName) .. " buff."
		end,
		arguments = {buffName = "text"}
	},
	["Raid Buffed"] = {
		test = function(checkTable) 
			if not UnitBuff("player", checkTable.buffName) then
				return false
			end
			for i = 1,40 do
				if UnitExists("raid"..i) and not UnitBuff("raid"..i, checkTable.buffName) then
					return false
				end				
			end
			return true
		end,
		text = function(checkTable) 
			return "all raid members have the " .. tostring(checkTable.buffName) .. " buff."
		end,
		arguments = {["buffName"] = "text"}
	},
	["Raid Buffed By Player"] = {
		test = function(checkTable)
			local name, _, _, _, _, _, _, unit = UnitBuff("player", checkTable.buffName) 
			if not name or not UnitIsUnit(unit,"player") then
				return false
			end	
			for i = 1,40 do
				if UnitExists("raid"..i) then
					local name, _, _, _, _, _, _, unit = UnitBuff("raid"..i, checkTable.buffName) 
					if not name or not UnitIsUnit(unit,"player") then
						return false
					end				
				end
			end
			return true
		end,
		text = function(checkTable) 
			return "all raid members have player's " .. tostring(checkTable.buffName) .. " buff."
		end,
		arguments = {["buffName"] = "text"}
	},
}

function WhatsNext:PrintChecks()
	for checkType,checkEntry in pairs(WhatsNext.defaultChecks) do
		print(checkType)
		print("   ", checkEntry.text({}))
	end
end

function WhatsNext:ResetTestChecks()
	checkResults = {}
end

function WhatsNext:TestCheck(checkTable)
	if not checkResults[self:GetCheckText(checkTable)] then
		if checks[checkTable.type].test(checkTable) then
			checkResults[self:GetCheckText(checkTable)] = checkTable.invert and 0 or 1
		else
			checkResults[self:GetCheckText(checkTable)] = checkTable.invert and 1 or 0
		end
	end
	return checkResults[self:GetCheckText(checkTable)] == 1
end

function WhatsNext:TestChecks(checkList)
	for i,checkTable in ipairs(checkList or {}) do
		if not self:TestCheck(checkTable) then
			return false
		end
	end
	return true
end

function WhatsNext:GetArguments(checkType)
	return checks[checkType].arguments
end

function WhatsNext:GetCheckText(checkTable)
	if not checks[checkTable.type] then
		print(checkTable.type)
	end
	return (checkTable.invert and "Not if " or "If ") .. checks[checkTable.type].text(checkTable)
end

function WhatsNext:GetSampleCheckTexts()
	local sampleTexts = {}
	for checkType, check in pairs(checks) do
		sampleTexts[checkType] = self:GetSampleCheckText(checkType)
	end
	return sampleTexts
end


function WhatsNext:GetSampleCheckText(checkType)
	local check = checks[checkType]
	local checkTable = {type = checkType}
	for var,varType in pairs(check.arguments) do
		checkTable[var] = "{"..var.."}"
	end
	return  "If " .. check.text(checkTable)
end

-------------------
-- Set up events
-------------------

local events = {}
function events:PLAYER_REGEN_ENABLED() 
	inCombat = false
	enemyCount = 0
end

function events:PLAYER_REGEN_DISABLED() 
	inCombat = true
	enemyCount = 1
end

function events:COMBAT_LOG_EVENT_UNFILTERED(timeStamp,event,hideCaster,sourceGUID,sourceName,sourceFlags,sourceRaidFlags,destGUID,destName,destFlags,destRaidFlags,spellID,spellName) 
	if UnitGUID("player") == destGUID then
		if event == "SWING_DAMAGE" or event == "SWING_MISSED" or event == "RANGE_DAMAGE" or event == "RANGE_MISSED" then
			lastBlockableAttackTaken = GetTime()
		elseif event == "SPELL_DAMAGE" or event == "SPELL_PERIODIC_DAMAGE" then
			lastUnblockableAttackTaken = GetTime()
		elseif event == "SPELL_MISSED" or event == "SPELL_PERIODIC_MISSED" then
			lastUnblockableAttackTaken = GetTime()
		end
	elseif UnitGUID("player") == sourceGUID then
		if spellName == SPELL_THUNDER_CLAP or spellName == SPELL_CLEAVE or spellName == SPELL_REVENGE or spellName == SPELL_SHOCKWAVE then
			if event == "SPELL_DAMAGE" then
				enemyCount = enemyCount + 1
			elseif event == "SPELL_CAST_SUCCESS" then
				enemyCount = 0
			end
		end
	end
end

local function EventFunc(self, event, ...)
	events[event](event, ...)
end
local eventHandler = CreateFrame("Frame", nil, WorldFrame)
eventHandler:SetScript("OnEvent", EventFunc)
for eventname in pairs(events) do eventHandler:RegisterEvent(eventname) end	